from urllib.parse import urlparse

import htmlgenerator as hg
from django.conf import settings
from django.core.signing import SignatureExpired, TimestampSigner
from django.http import HttpResponseNotFound, HttpResponseRedirect
from django.http.request import QueryDict
from django.urls import path, resolve
from django.utils.translation import gettext as _
from guardian.utils import get_anonymous_user

from basxbread import layout, menu, views
from basxbread.utils import ModelHref, aslayout, default_model_paths

from . import models


@aslayout
def publicurlview(request, token):
    # unpack and verify token
    if ":" not in token:
        return HttpResponseNotFound()
    token, salt = token.rsplit(":", 1)
    signer = TimestampSigner(salt=salt)
    pk = int(signer.unsign(token))
    url = models.PublicURL.objects.get(pk=pk)
    if url.valid_for is not None:
        try:
            pk = int(signer.unsign(token, max_age=url.valid_for))
        except SignatureExpired:
            return hg.DIV(
                _("Link has expired"), style="margin: auto auto; text-align: center"
            )

    # call actual view
    parsed = urlparse(url.url)
    match_ = resolve(parsed.path)
    request.resolver_match = match_  # make sure view uses correct resolver_match
    request.user = get_anonymous_user()
    request.GET = QueryDict(parsed.query, mutable=True)
    request.GET[settings.HIDEMENUS_URLPARAMETER] = True
    response = match_.func(request, *match_.args, **match_.kwargs)  # call view function

    # check if a form has successfully been submited, if so return a thank you page
    if (
        request.method == "POST"
        and url.has_form
        and isinstance(response, HttpResponseRedirect)
    ):
        return hg.DIV(
            hg.H2(_("Data has been submitted")),
            hg.H4(url.thankyou_text),
            hg.A(_("Back"), href=request.get_full_path()),
            hg.SCRIPT(
                "if (window.history.replaceState)"
                "{window.history.replaceState(null, null, window.location.href);}"
            ),
            style="margin: auto auto; text-align: center",
        )
    return response


urlpatterns = [
    path("public/<str:token>", publicurlview, name="publicurl"),
    *default_model_paths(
        models.PublicURL,
        browseview=views.BrowseView._with(
            columns=[
                "name",
                "url",
                layout.datatable.DataTableColumn(
                    _("Public URL"),
                    hg.F(
                        lambda c: hg.SPAN(
                            c["row"].absolute_publicurl(c["request"]),
                            style="word-break: break-all",
                        )
                    ),
                ),
                layout.datatable.DataTableColumn(
                    "",
                    hg.F(
                        lambda c: layout.button.Button(
                            icon="copy",
                            buttontype="ghost",
                            small=True,
                            onclick=f"navigator.clipboard.writeText('{c['row'].absolute_publicurl(c['request'])}');",
                        )
                    ),
                ),
                "created",
                "valid_for",
            ],
            rowactions=(views.BrowseView.editlink(), views.BrowseView.deletelink()),
        ),
    ),
]

menu.registeritem(
    menu.Item(
        group=models.PublicURL._meta.verbose_name_plural,
        link=menu.Link(
            href=ModelHref(models.PublicURL, "browse"),
            label=models.PublicURL._meta.verbose_name_plural,
            iconname="link",
        ),
    )
)



from secrets import token_hex
from urllib.parse import urlparse

from django.core.exceptions import ValidationError
from django.core.signing import TimestampSigner
from django.db import models
from django.urls import Resolver404, resolve, reverse
from django.utils.translation import gettext as _


def validate_url(value):
    try:
        resolve(urlparse(value).path)
    except Resolver404:
        raise ValidationError(
            _("%(value)s is not a valid internal URL"),
            params={"value": value},
        )


class PublicURL(models.Model):
    name = models.CharField(_("Name"), max_length=255)
    url = models.CharField(
        _("URL"),
        max_length=2048,
        validators=[validate_url],
        help_text=_("Internal URL"),
    )
    salt = models.CharField(_("Salt"), max_length=32, editable=False)
    created = models.DateTimeField(_("Created"), auto_now_add=True, editable=False)
    valid_for = models.DurationField(_("Valid for"), null=True, blank=True)
    has_form = models.BooleanField(_("Has form"), default=False)
    thankyou_text = models.TextField(
        _("Thank-you text"),
        default=_("Thank you"),
        help_text=_("Will be shown if a form has successfully been submitted."),
    )

    def publicurl(self):
        return reverse(
            "publicurl",
            kwargs={
                "token": f"{TimestampSigner(salt=self.salt).sign(self.pk)}:{self.salt}"
            },
        )

    setattr(publicurl, "verbose_name", _("Public URL"))

    def absolute_publicurl(self, request):
        return request.build_absolute_uri(location=self.publicurl())

    def save(self, *args, **kwargs):
        if not self.pk:  # prevent to resuse salt when object is copied
            self.salt = None
        if not self.salt:
            self.salt = token_hex(16)
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _("Public URL")
        verbose_name_plural = _("Public URLs")
        ordering = ["-created"]



# Generated by Django 4.0.5 on 2022-08-22 03:43

from typing import Any, List

import django.db.models.deletion
from django.db import connection, migrations, models


class Migration(migrations.Migration):
    def copy_existing(apps, _):
        if not apps.is_installed("basxconnect.core"):
            return
        with connection.cursor() as cursor:
            cursor.execute(
                "INSERT INTO taxonomy_vocabulary SELECT * FROM core_vocabulary;"
            )
            cursor.execute("INSERT INTO taxonomy_term SELECT * FROM core_term;")

    initial = True

    dependencies: List[Any] = []

    operations = [
        migrations.CreateModel(
            name="Vocabulary",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(max_length=255, unique=True, verbose_name="Name"),
                ),
                (
                    "slug",
                    models.SlugField(
                        help_text="slug is human-readable, to make referencing easier",
                        unique=True,
                        verbose_name="Slug",
                    ),
                ),
            ],
            options={
                "verbose_name": "Vocabulary",
                "verbose_name_plural": "Vocabularies",
                "ordering": ["name"],
            },
        ),
        migrations.CreateModel(
            name="Term",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("term", models.CharField(max_length=255, verbose_name="Term")),
                (
                    "slug",
                    models.CharField(
                        blank=True, max_length=255, unique=True, verbose_name="Slug"
                    ),
                ),
                (
                    "vocabulary",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="taxonomy.vocabulary",
                    ),
                ),
            ],
            options={
                "verbose_name": "Term",
                "verbose_name_plural": "Terms",
                "order_with_respect_to": "vocabulary",
            },
        ),
    ]



import htmlgenerator as hg
from django.forms import formset_factory
from django.urls import path

from basxbread import layout
from basxbread.layout.components import forms
from basxbread.utils import aslayout

from .parser import FieldForm, ModelForm, parse


@aslayout
def test(request):

    with open(
        "/home/sam/p/pythonpackages/bread/basxbread/contrib/reports/models.py"
    ) as f:
        print(parse(f.read()))

    modelforms = {
        "a1": [ModelForm(prefix="a1"), formset_factory(FieldForm)(prefix="a1")],
        "a2": [ModelForm(prefix="a2"), formset_factory(FieldForm)(prefix="a2")],
    }
    return hg.DIV(
        hg.H1("ModelEdit test page"),
        hg.Iterator(
            modelforms.keys(),
            "modelname",
            layout.button.Button(
                hg.C("modelname"),
                onclick=hg.format(
                    "$$('.modelform')._.style({{ 'display': 'none' }}) ; $('#model-{}')._.style({{'display': 'block'}})",
                    hg.C("modelname"),
                ),
            ),
        ),
        hg.Iterator(
            modelforms.items(),
            "modelform",
            hg.DIV(
                forms.Form(
                    hg.C("modelform")[1][0],
                    *[forms.FormField(f) for f in ModelForm().fields]
                ),
                forms.Formset.as_plain(
                    hg.C("modelform")[1][1],
                    content=hg.DIV(
                        *[forms.FormField(f) for f in FieldForm().fields],
                        style="background-color: #fff"
                    ),
                ),
                id=hg.format("model-{}", hg.C("modelform")[0]),
                style="display: none",
                _class="modelform",
            ),
        ),
    )


urlpatterns = [
    path("test", test),
]



from django.apps import AppConfig
from django.contrib.messages.constants import DEFAULT_TAGS
from django.utils.translation import gettext as _


class BasxBreadConfig(AppConfig):

    name = "basxbread"
    verbose_name = "basx BREAD Engine"

    def ready(self):
        # trigger translation of message tags
        [_(tag.capitalize()) for tag in DEFAULT_TAGS.values()]
        patch_django_filters_verbose_name_func()


def patch_django_filters_verbose_name_func():
    # django filters verbose_field_name function is not great at using the
    # verbose names on related fields, patching this here
    from django_filters import utils
    from django_filters.utils import ForeignObjectRel, force_str, get_field_parts

    def improved_verbose_field_name_func(model, field_name):
        if field_name is None:
            return "[invalid name]"

        parts = get_field_parts(model, field_name)
        if not parts:
            return "[invalid name]"

        names = []
        for part in parts:
            if isinstance(part, ForeignObjectRel):
                if hasattr(part, "related_model"):
                    names.append(force_str(part.related_model._meta.verbose_name))
                else:
                    return "[invalid name]"
            else:
                names.append(force_str(part.verbose_name))

        return " ".join(names)

    utils.verbose_field_name = improved_verbose_field_name_func



# Generated by Django 4.0.5 on 2022-08-24 07:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reports', '0010_alter_reportcolumn_sortingname'),
    ]

    operations = [
        migrations.AddField(
            model_name='reportcolumn',
            name='cell_template',
            field=models.TextField(blank=True, help_text="Jinja template with 'value' in context", verbose_name='Cell template'),
        ),
    ]



# Generated by Django 4.0.5 on 2022-07-16 06:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('publicurls', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='publicurl',
            options={'ordering': ['-created'], 'verbose_name': 'Public URL', 'verbose_name_plural': 'Public URLs'},
        ),
        migrations.AddField(
            model_name='publicurl',
            name='thankyou_text',
            field=models.TextField(default='Thank you', help_text='Will be shown if a form has successfully been submitted.', verbose_name='Thank-you text'),
        ),
    ]



# Generated by Django 4.0.5 on 2022-07-15 09:56

from typing import List

from django.db import migrations, models

import basxbread.contrib.publicurls.models


class Migration(migrations.Migration):

    initial = True

    dependencies: List[None] = []

    operations = [
        migrations.CreateModel(
            name="PublicURL",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
                (
                    "url",
                    models.CharField(
                        max_length=2048,
                        validators=[basxbread.contrib.publicurls.models.validate_url],
                        verbose_name="URL",
                    ),
                ),
                (
                    "salt",
                    models.CharField(
                        editable=False, max_length=32, verbose_name="Salt"
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created"),
                ),
                (
                    "valid_for",
                    models.DurationField(
                        blank=True, null=True, verbose_name="Valid for"
                    ),
                ),
                (
                    "has_form",
                    models.BooleanField(default=False, verbose_name="Has form"),
                ),
            ],
            options={
                "verbose_name": "Public URL",
                "verbose_name_plural": "Public URLs",
            },
        ),
    ]



import urllib

from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from django.views.generic import RedirectView
from guardian.mixins import PermissionRequiredMixin

from ..utils import reverse_model
from .util import BaseView


class DeleteView(BaseView, PermissionRequiredMixin, RedirectView):
    """TODO: documentation"""

    model = None
    softdeletefield = None  # set to a boolean field on the modle which will be set to True instead of deleting the object
    accept_global_perms = True
    urlparams = (("pk", int),)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # raise exception in case the model has no field with the given name
        if self.softdeletefield:
            self.model._meta.get_field(self.softdeletefield)

    def get_required_permissions(self, request):
        if "restore" in self.request.GET:
            return [
                f"{self.model._meta.app_label}.change_{self.model.__name__.lower()}"
            ]
        return [f"{self.model._meta.app_label}.delete_{self.model.__name__.lower()}"]

    def post(self, *args, **kwargs):
        instance = get_object_or_404(self.model, pk=self.kwargs.get("pk"))
        restore = "restore" in self.request.GET
        if self.softdeletefield:
            setattr(instance, self.softdeletefield, not restore)
            instance.save()
            msg = (
                _("Restored %(modelname)s %(objectname)s")
                if restore
                else _("Deleted %(modelname)s %(objectname)s")
            )
        else:
            instance.delete()
            msg = _("Deleted %(modelname)s %(objectname)s")
        messages.success(
            self.request,
            msg
            % {
                "objectname": instance,
                "modelname": self.model._meta.verbose_name,
            },
        )
        ret = super().post(*args, **kwargs)
        if self.ajax_urlparameter in self.request.GET:
            ret = HttpResponse("OK")
            # This header will be processed by htmx
            # in order to reload the whole page automatically
            ret["HX-Refresh"] = "true"
        return ret

    def get_redirect_url(self, *args, **kwargs):
        if self.request.GET.get("next"):
            return urllib.parse.unquote(self.request.GET["next"])
        if "restore" in self.request.GET and self.softdeletefield:
            return reverse_model(
                self.model, "read", kwargs={"pk": self.kwargs.get("pk")}
            )
        return reverse_model(self.model, "browse")



from typing import Callable, Iterable, List, NamedTuple, Optional, Union

import django_filters
import htmlgenerator as hg
from django import forms
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy
from django.views.generic import ListView
from djangoql.exceptions import DjangoQLError
from djangoql.queryset import apply_search
from guardian.mixins import PermissionListMixin

from .. import layout
from ..utils import (
    Link,
    ModelHref,
    expand_ALL_constant,
    filter_fieldlist,
    generate_excel,
    link_with_urlparameters,
    queryset_from_fields,
    xlsxresponse,
)
from .util import BaseView


class BulkAction(NamedTuple):
    name: str
    label: str
    action: Callable[[HttpRequest, models.query.QuerySet], Optional[HttpResponse]]
    iconname: str = "fade"
    permissions: List[str] = []

    def has_permission(self, request, obj=None):
        return all(
            [
                request.user.has_perm(perm, obj) or request.user.has_perm(perm)
                for perm in self.permissions
            ]
        )


def default_bulkactions(model, columns=["__all__"]):
    return (
        BulkAction(
            "excel",
            label=_("Excel"),
            iconname="download",
            action=lambda request, qs: export(qs, columns),
            permissions=[f"{model._meta.app_label}.view_{model._meta.model_name}"],
        ),
        BulkAction(
            "delete",
            label=_("Delete"),
            iconname="trash-can",
            action=delete,
            permissions=[f"{model._meta.app_label}.add_{model._meta.model_name}"],
        ),
    )


def order_queryset_by_urlparameter(qs, order):
    """Can used to order a queryset by a user-provided string, e.g. through a GET query parameter"""
    if order:
        if order.endswith("__int"):
            order = order[: -len("__int")]
            qs = qs.order_by(
                models.functions.Cast(order[1:], models.IntegerField()).desc()
                if order.startswith("-")
                else models.functions.Cast(order, models.IntegerField())
            )
        else:
            qs = qs.order_by(
                models.functions.Lower(order[1:]).desc()
                if order.startswith("-")
                else models.functions.Lower(order)
            )
    return qs


class BrowseView(BaseView, LoginRequiredMixin, PermissionListMixin, ListView):
    """TODO: documentation"""

    orderingurlparameter: str = "ordering"

    # see basxbread/static/js/main.js:submitbulkaction and basxbread/layout/components/datatable.py
    objectids_urlparameter: str = "_selected"

    bulkaction_urlparameter: str = "_bulkaction"
    items_per_page_options: Optional[Iterable[int]] = None
    itemsperpage_urlparameter: str = "itemsperpage"
    search_urlparameter: str = "q"

    title: Union[hg.BaseElement, str] = ""
    columns: Iterable[Union[str, layout.datatable.DataTableColumn]] = ("__all__",)
    rowclickaction: Optional[Link] = None
    filterconfig: Optional[tuple] = None
    filterset_class: Optional[django_filters.FilterSet] = None

    # bulkactions: List[(Link, function(request, queryset))]
    # - link.js should be a slug and not a URL
    # - if the function returns a HttpResponse, the response is returned
    #   instead of the browse view result
    bulkactions: Iterable[
        Union[Link, Callable[[HttpRequest, models.QuerySet], Union[None, HttpResponse]]]
    ] = ()

    rowactions: Iterable[Link] = ()  # list of links
    backurl = None
    primary_button = None

    # if set will be used to save the state of the url parameters and restore them on the next call
    viewstate_sessionkey: Optional[str] = None

    def __init__(self, *args, **kwargs):
        self.orderingurlparameter = (
            kwargs.get("orderingurlparameter") or self.orderingurlparameter
        )
        self.itemsperpage_urlparameter = (
            kwargs.get("itemsperpage_urlparameter") or self.itemsperpage_urlparameter
        )
        self.objectids_urlparameter = (
            kwargs.get("objectids_urlparameter") or self.objectids_urlparameter
        )
        self.bulkaction_urlparameter = (
            kwargs.get("bulkaction_urlparameter") or self.bulkaction_urlparameter
        )
        self.items_per_page_options = (
            kwargs.get("items_per_page_options")
            or self.items_per_page_options
            or getattr(settings, "DEFAULT_PAGINATION_CHOICES")
        )
        self.search_urlparameter = (
            kwargs.get("search_urlparameter") or self.search_urlparameter
        )
        self.title = kwargs.get("title") or self.title
        self.rowactions = kwargs.get("rowactions") or self.rowactions
        self.model = kwargs.get("model") or self.model
        self.columns = expand_ALL_constant(
            self.model, kwargs.get("columns") or self.columns
        )
        self.rowclickaction = kwargs.get("rowclickaction") or self.rowclickaction
        self.backurl = kwargs.get("backurl") or self.backurl
        self.primary_button = kwargs.get("primary_button") or self.primary_button
        self.viewstate_sessionkey = (
            kwargs.get("viewstate_sessionkey") or self.viewstate_sessionkey
        )
        self.filterconfig = kwargs.get("filterconfig") or self.filterconfig
        self.filterset_class = self.filterset_class or parse_filterconfig(
            self.model,
            self.filterconfig
            or (
                ("AND",)
                + tuple(
                    get_filtername(column)
                    for column in self.columns
                    if get_filtername(column)
                )
            ),
            prefix="f",
        )
        super().__init__(*args, **kwargs)
        self.bulkactions = (
            kwargs.get("bulkactions")
            or self.bulkactions
            or default_bulkactions(self.model, self.columns)
        )

    def get_layout(self, **datatable_kwargs):
        qs = self.get_queryset()
        # re-mapping the Links because the URL is not supposed to be a real URL but an identifier
        # for the bulk action
        # TODO: This is a bit ugly but we can reuse the Link type for icon, label and permissions
        bulkactions = [
            Link(
                link_with_urlparameters(
                    self.request, **{self.bulkaction_urlparameter: action.name}
                ),
                label=action.label,
                iconname=action.iconname,
            )
            for action in self.bulkactions
            if action.has_permission(self.request)
        ]
        qs = self.get_queryset()
        paginate_by = self.get_paginate_by(qs)
        if paginate_by is None:
            paginate_by = qs.count()
        qs = self.paginate_queryset(qs, paginate_by)[2] if paginate_by > 0 else qs

        return layout.datatable.DataTable.from_queryset(
            queryset=qs,
            columns=self.columns,
            bulkactions=bulkactions,
            rowactions=self.rowactions,
            rowactions_dropdown=len(self.rowactions)
            > 2,  # recommendation from carbon design
            rowclickaction=self.rowclickaction,
            pagination_config=layout.pagination.PaginationConfig(
                items_per_page_options=self.items_per_page_options,
                page_urlparameter=self.page_kwarg,
                paginator=self.get_paginator(qs, paginate_by),
                itemsperpage_urlparameter=self.itemsperpage_urlparameter,
            )
            if paginate_by > 0
            else None,
            checkbox_for_bulkaction_name=self.objectids_urlparameter,
            title=self.title,
            settingspanel=self.get_settingspanel(),
            backurl=self.backurl,
            primary_button=self.primary_button,
            search_urlparameter=self.search_urlparameter,
            **datatable_kwargs,
        )

    def get_context_data(self, *args, **kwargs):
        return {
            **super().get_context_data(*args, **kwargs),
            "layout": self._get_layout_cached(),
            "pagetitle": self.title or self.model._meta.verbose_name_plural,
        }

    def get_settingspanel(self):
        return build_filterpanel(
            self.filterset_class(self.request.GET, queryset=self.get_queryset())
        )

    def get_required_permissions(self, request):
        return [f"{self.model._meta.app_label}.view_{self.model.__name__.lower()}"]

    def get(self, *args, **kwargs):
        if "reset" in self.request.GET:
            if (
                self.viewstate_sessionkey
                and self.viewstate_sessionkey in self.request.session
            ):
                del self.request.session[self.viewstate_sessionkey]
            return redirect(self.request.path)
        if self.bulkaction_urlparameter in self.request.GET:
            bulkactions = {
                action.name: action.action
                for action in self.bulkactions
                if action.has_permission(self.request)
            }
            if self.request.GET[self.bulkaction_urlparameter] not in bulkactions:
                messages.error(
                    self.request,
                    _("Action '%s' is not configured for this view")
                    % self.request.GET[self.bulkaction_urlparameter],
                )
            else:
                ret = bulkactions[self.request.GET[self.bulkaction_urlparameter]](
                    self.request, self.get_queryset()
                )
                params = self.request.GET.copy()
                del params[self.bulkaction_urlparameter]
                del params[self.objectids_urlparameter]
                if ret is None:
                    return redirect(self.request.path + "?" + params.urlencode())
                return ret
        # for normal GET requests save query if saving state is enabled or reload last state
        if self.viewstate_sessionkey:
            if not self.request.GET and self.request.session.get(
                self.viewstate_sessionkey, None
            ):
                return redirect(
                    self.request.path
                    + "?"
                    + self.request.session[self.viewstate_sessionkey]
                )
            self.request.session[
                self.viewstate_sessionkey
            ] = self.request.GET.urlencode()

        return super().get(*args, **kwargs)

    def get_paginate_by(self, queryset):
        ret = self.request.GET.get(
            self.itemsperpage_urlparameter, self.items_per_page_options[0]
        )
        if str(ret) == "-1":
            return None

        return int(ret)

    def get_queryset(self):
        """Prefetch related tables to speed up queries. Also order result by get-parameters."""
        qs = super().get_queryset()
        if self.search_urlparameter and self.search_urlparameter in self.request.GET:
            searchquery = self.request.GET[self.search_urlparameter].strip()
            if searchquery.startswith("="):
                try:
                    qs = apply_search(qs, searchquery[1:])
                except DjangoQLError as e:
                    messages.error(
                        self.request,
                        _("Bad filter string '%s': '%s'") % (searchquery, e),
                    )

            else:
                qs = self.model.objects.filter(
                    queryset_from_fields.get_field_queryset(
                        [*self.model._meta.fields, *self.model._meta.many_to_many],
                        searchquery,
                    )
                )

        filterset = self.filterset_class(self.request.GET, queryset=qs)
        qs = filterset.qs

        selectedobjects = self.request.GET.getlist(self.objectids_urlparameter)
        if selectedobjects and "all" not in selectedobjects:
            qs &= super().get_queryset().filter(pk__in=selectedobjects)

        qs = order_queryset_by_urlparameter(
            qs, self.request.GET.get(self.orderingurlparameter)
        )
        return qs

    @staticmethod
    def gen_rowclickaction(modelaction, **kwargs):
        """
        Shortcut to get a Link to a model view.
        The default models views in basxbread are "read", "edit", "delete".
        :param modelaction: A model view whose name has been generated
                            with ``basxbread.utils.urls.model_urlname``
        """
        return Link(
            label="",
            href=ModelHref.from_object(hg.C("row"), modelaction, **kwargs),
            iconname=None,
        )

    @staticmethod
    def editlink(return_to_current=True, **attributes):
        return Link(
            href=ModelHref.from_object(
                hg.C("row"), "edit", return_to_current=return_to_current
            ),
            label=_("Edit"),
            iconname="edit",
            attributes=attributes,
        )

    @staticmethod
    def deletelink(return_to_current=True):
        return Link(
            href=ModelHref.from_object(
                hg.C("row"), "delete", return_to_current=return_to_current
            ),
            label=_("Delete"),
            iconname="delete",
        )


# helper function to export a queryset to excel
def export(queryset, columns):
    if "__all__" in columns:
        columns = filter_fieldlist(queryset.model, columns)
    columndefinitions = {}
    for column in columns:
        if not (
            isinstance(column, layout.datatable.DataTableColumn)
            or isinstance(column, str)
        ):
            raise ValueError(
                "Argument 'columns' needs to be of a list with items of type str "
                f"or DataTableColumn, but found {column}"
            )
        if isinstance(column, str):
            column = layout.datatable.DataTableColumn(
                layout.ObjectFieldLabel(column, "model"),
                layout.ObjectFieldValue(column, "row"),
            )

        columndefinitions[
            hg.render(hg.BaseElement(column.header), {"model": queryset.model})
        ] = lambda row, column=column: hg.render(
            hg.BaseElement(column.cell), {"row": row}
        )

    workbook = generate_excel(queryset, columndefinitions)
    workbook.title = queryset.model._meta.verbose_name
    return xlsxresponse(workbook, workbook.title)


def delete(request, queryset, softdeletefield=None, required_permissions=None):
    if required_permissions is None:
        required_permissions = [
            f"{queryset.model._meta.app_label}.delete_{queryset.model.__name__.lower()}"
        ]

    deleted = 0
    for instance in queryset:
        try:
            if not request.user.has_perm(required_permissions, instance):
                # we throw an exception here because the user not supposed to
                # see the option to delete an object anyway, if he does not have the permssions
                # the queryset should already be filtered
                raise Exception(
                    _("Your user has not the permissions to delete %s") % instance
                )
            if softdeletefield:
                if not getattr(instance, softdeletefield):
                    setattr(instance, softdeletefield, True)
                    instance.save()
                    deleted += 1
            else:
                instance.delete()
                deleted += 1
        except Exception as e:
            messages.error(
                request,
                _("%s could not be deleted: %s") % (object, e),
            )

    messages.success(
        request,
        _("Deleted %(count)s %(modelname)s")
        % {
            "count": deleted,
            "modelname": queryset.model._meta.verbose_name_plural
            if deleted > 1
            else queryset.model._meta.verbose_name,
        },
    )


def restore(request, queryset, softdeletefield, required_permissions=None):
    if required_permissions is None:
        required_permissions = [
            f"{queryset.model._meta.app_label}.change_{queryset.model.__name__.lower()}"
        ]

    restored = 0
    for instance in queryset:
        try:
            if not request.user.has_perm(required_permissions, instance):
                # we throw an exception here because the user not supposed to
                # see the option to restore an object anyway, if he does not have the permssions
                # the queryset should already be filtered
                raise Exception(
                    _("Your user has not the permissions to restore %s") % instance
                )
            if getattr(instance, softdeletefield, False):
                setattr(instance, softdeletefield, False)
                instance.save()
                restored += 1
        except Exception as e:
            messages.error(
                request,
                _("%s could not be restored: %s") % (object, e),
            )

    messages.success(
        request,
        _("Restored %(count)s %(modelname)s")
        % {
            "count": restored,
            "modelname": queryset.model._meta.verbose_name_plural
            if restored > 1
            else queryset.model._meta.verbose_name,
        },
    )


class AndGroup(django_filters.FilterSet):
    prefix = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subgroups = []
        for group in self.subgroup_classes:
            self.subgroups.append(group(*args, **kwargs))
        self.form_prefix = self.prefix

    def filter_queryset(self, queryset):
        for name, value in self.form.cleaned_data.items():
            queryset = self.filters[name].filter(queryset, value)
            assert isinstance(
                queryset, models.QuerySet
            ), "Expected '%s.%s' to return a QuerySet, but got a %s instead." % (
                type(self).__name__,
                name,
                type(queryset).__name__,
            )

        for group in self.subgroups:
            queryset = group.filter_queryset(queryset)

        return queryset

    @property
    def errors(self):
        ret = django_filters.FilterSet.errors.fget(self)
        for subgroup in self.subgroups:
            ret.update(subgroup.errors)
        return ret


class OrGroup(django_filters.FilterSet):
    prefix = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subgroups = []
        for group in self.subgroup_classes:
            self.subgroups.append(group(*args, **kwargs))
        self.form_prefix = self.prefix

    def filter_queryset(self, queryset):
        if (
            len([i for i in self.subgroups if i.form.has_changed()]) == 0
            and len(self.form.cleaned_data) == 0
        ):
            return queryset
        basefilterset = queryset.none()
        for name, value in self.form.cleaned_data.items():
            basefilterset |= self.filters[name].filter(queryset, value)
            assert isinstance(
                basefilterset, models.QuerySet
            ), "Expected '%s.%s' to return a QuerySet, but got a %s instead." % (
                type(self).__name__,
                name,
                type(basefilterset).__name__,
            )
        for group in self.subgroups:
            if group.form.has_changed():  # ignore subgroup with no changes
                basefilterset |= group.filter_queryset(queryset)
        return basefilterset

    @property
    def errors(self):
        ret = django_filters.FilterSet.errors.fget(self)
        for subgroup in self.subgroups:
            ret.update(subgroup.errors)
        return ret


def parse_filterconfig(basemodel, filterconfig, prefix):
    """
    filterconfig: Tree in the form of
                  ("and"
                      ("or", fieldname1, fieldname2),
                      ("or", fieldname3, fieldname4),
                  )
    """
    FILTERSETTYPE = {"and": AndGroup, "or": OrGroup}
    grouptype, *subfields = filterconfig
    if grouptype.lower() not in FILTERSETTYPE:
        raise ValueError(
            f"Invalid filter group specified '{filterconfig[0]}', must be 'and' or 'or'"
        )

    fields = []
    subgroups = []
    n = 0
    for f in subfields:
        if isinstance(f, str):
            try:
                modelfield = get_field(basemodel, f)
            except FieldDoesNotExist:
                continue
            # ignore filefields
            if not isinstance(
                modelfield, (models.FileField, GenericForeignKey, GenericRelation)
            ):
                fields.append(f)
        elif isinstance(f, Iterable):
            subgroups.append(parse_filterconfig(basemodel, f, prefix + str(n)))
            n += 1
        else:
            raise ValueError(
                f"Declared filter field '{f}' is not of type {(str, OrGroup, AndGroup)} but {type(f)}"
            )

    meta = type(
        "Meta",
        (),
        {
            "model": basemodel,
            "fields": fields,
            "filter_overrides": FILTER_OVERRIDES,
        },
    )

    ret = type(
        f"{basemodel.__name__}FilterSet",
        (FILTERSETTYPE[grouptype.lower()],),
        {"Meta": meta, "subgroup_classes": subgroups, "prefix": f"filter_{prefix}"},
    )
    return ret


def get_filtername(column):
    if isinstance(column, str):
        return column.replace(".", LOOKUP_SEP)
    if isinstance(column, layout.datatable.DataTableColumn):
        return column.filtername
    return None


def get_field(basemodel, fieldname):
    if LOOKUP_SEP in fieldname:
        relatedfield, rest = fieldname.split(LOOKUP_SEP, 1)
        return get_field(basemodel._meta.get_field(relatedfield).related_model, rest)
    return basemodel._meta.get_field(fieldname)


def build_filterpanel(filterset):
    return hg.DIV(
        hg.DIV(
            hg.DIV(
                hg.DIV(_("Filter"), style="margin-bottom: 1rem"),
                hg.DIV(
                    layout.components.forms.Form(
                        forms.Form(),
                        _build_filter_ui_recursive(filterset),
                        _class="filterform",
                        method="GET",
                    ),
                    style="display: flex",
                ),
            ),
            style="display: flex; padding: 24px 32px 0 32px",
        ),
        hg.DIV(
            layout.button.Button(
                _("Cancel"),
                buttontype="ghost",
                onclick="this.closest('.settingscontainer').style.display = 'none'",
            ),
            layout.button.Button.from_link(
                Link(
                    label=_("Reset"),
                    href=hg.format("{}?reset=1", hg.C("request").path),
                    iconname=None,
                ),
                buttontype="secondary",
            ),
            layout.button.Button(
                pgettext_lazy("apply filter", "Filter"),
                onclick="""this.closest('.filterpanel').querySelector('.filterform').submit();""",
            ),
            style="display: flex; justify-content: flex-end; margin-top: 24px",
            _class="bx--modal-footer",
        ),
        _class="filterpanel",
        style="background-color: #fff",
    )


def _build_filter_ui_recursive(filterset):
    flexdir = ""
    if isinstance(filterset, AndGroup):
        flexdir = "column"
    elif isinstance(filterset, OrGroup):
        flexdir = "row"

    return hg.DIV(
        layout.components.forms.Form(
            filterset.form,
            *[
                layout.components.forms.FormField(
                    f, no_wrapper=True, style="margin-bottom: 1rem; margin-right: 1rem"
                )
                for f in filterset.form.fields
            ],
            standalone=False,
            *[_build_filter_ui_recursive(f) for f in filterset.subgroups],
        ),
        style="display: flex; flex-direction: " + flexdir,
    )


FILTER_OVERRIDES = {
    models.CharField: {
        "filter_class": django_filters.CharFilter,
        "extra": lambda f: {"lookup_expr": "icontains"},
    },
    models.TextField: {
        "filter_class": django_filters.CharFilter,
        "extra": lambda f: {"lookup_expr": "icontains"},
    },
    models.EmailField: {
        "filter_class": django_filters.CharFilter,
        "extra": lambda f: {"lookup_expr": "icontains"},
    },
    models.URLField: {
        "filter_class": django_filters.CharFilter,
        "extra": lambda f: {"lookup_expr": "icontains"},
    },
    models.DateField: {
        "filter_class": django_filters.DateFromToRangeFilter,
        "extra": lambda f: {
            "widget": django_filters.widgets.DateRangeWidget(
                attrs={"type": "text", "class": "validate datepicker"}
            )
        },
    },
    models.DateTimeField: {
        "filter_class": django_filters.DateFromToRangeFilter,
        "extra": lambda f: {
            "widget": django_filters.widgets.DateRangeWidget(
                attrs={"type": "text", "class": "validate datepicker"}
            )
        },
    },
}



# Generated by Django 3.2.6 on 2021-08-26 12:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("reports", "0004_auto_20210619_1502"),
    ]

    operations = [
        migrations.AlterField(
            model_name="report",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="reportcolumn",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



import htmlgenerator as hg
from django import forms
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _
from jinja2.sandbox import SandboxedEnvironment

from basxbread import layout
from basxbread.querysetfield import QuerysetField, parsequeryexpression

from ...layout.components.datatable import DataTableColumn


def available_report_filters(modelfield, request, report):
    from django.conf import settings

    return tuple((i, i) for i in getattr(settings, "REPORT_FILTERS", {}).keys())


class Report(models.Model):
    created = models.DateField(_("Created"), auto_now_add=True)
    name = models.CharField(_("Name"), max_length=255)
    model = models.ForeignKey(
        ContentType, on_delete=models.PROTECT, verbose_name=_("Model")
    )
    model.formfield_kwargs = {
        "queryset": ContentType.objects.all().order_by("app_label", "model")
    }
    filter = QuerysetField(_("Filter"), modelfieldname="model")
    custom_queryset = models.CharField(
        _("Custom Filter"),
        max_length=255,
        help_text=_(
            "Key in 'settings.REPORT_FILTERS' must be a function returning a queryset"
        ),
        blank=True,
    )
    custom_queryset.lazy_choices = available_report_filters
    pagination = models.PositiveIntegerField(
        _("Pagination"),
        default=0,
        help_text=_(
            "How many items to display per page when viewing the "
            "report in the browser, 0 for everything on one page"
        ),
    )

    @property
    def preview(self):
        columns = []
        for column in self.columns.all():
            columns.append(
                DataTableColumn(
                    header=column.header,
                    cell=column.render_element("row"),
                )
            )
        qs = self.queryset
        if qs is None:
            return hg.BaseElement("Model does no longer exists!")

        return layout.datatable.DataTable.from_queryset(
            title=_("Preview"),
            helper_text="",
            queryset=qs[:10],
            columns=columns,
            primary_button="",
        )

    @property
    def queryset(self):
        if self.custom_queryset and self.custom_queryset in getattr(
            settings, "REPORT_FILTERS", {}
        ):
            # do not check whether the settings exists, an exception can be raised automatically
            ret = getattr(settings, "REPORT_FILTERS", {}).get(
                self.custom_queryset, lambda model: model.objects.none()
            )(self.model.model_class())
            if not isinstance(ret, models.QuerySet):
                raise ValueError(
                    _(
                        'settings.REPORT_FILTERS["%s"] did not return a queryset but returned: %s'
                    )
                    % (self.custom_queryset, ret)
                )
            if ret.model != self.model.model_class():
                raise ValueError(
                    _(
                        'settings.REPORT_FILTERS["%s"] did not return a queryset for %s but for %s'
                    )
                    % (self.custom_queryset, self.model.model_class(), ret.model)
                )
            return parsequeryexpression(ret, self.filter.raw).queryset
        return self.filter.queryset

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _("Report")
        verbose_name_plural = _("Reports")
        ordering = ["name"]


class ReportColumn(models.Model):
    AGGREGATIONS = {
        "count": "",
        "sum": "",
    }
    report = models.ForeignKey(Report, on_delete=models.CASCADE, related_name="columns")
    header = models.CharField(_("Header"), max_length=255)
    column = models.CharField(
        _("Column"), max_length=255, help_text=_("Value expression (see 'Help')")
    )
    cell_template = models.TextField(
        _("Cell template"),
        blank=True,
        help_text=hg.mark_safe(
            _(
                "Optional Jinja template with 'value' in context.<br/>Use e.g. {{ value }} to simply display the value."
            )
        ),
    )
    cell_template.formfield_kwargs = {"widget": forms.Textarea(attrs={"rows": 1})}
    sortingname = models.CharField(
        _("Sortingname"),
        max_length=255,
        blank=True,
        help_text=_("Django sorting expression"),
    )
    aggregation = models.CharField(
        _("Aggregation"), max_length=64, choices=tuple(AGGREGATIONS.items()), blank=True
    )

    def render_element(self, rowvariable: str) -> hg.BaseElement:
        if self.cell_template:

            def render(context):
                elementenv = SandboxedEnvironment()
                elementenv.filters["map"] = lambda value, map: map.get(value, value)
                try:
                    return hg.BaseElement(
                        elementenv.from_string(self.cell_template).render(
                            value=hg.resolve_lookup(context[rowvariable], self.column),
                        )
                    )
                except Exception as e:
                    return hg.BaseElement(f"### ERROR: {e} ###")

            return hg.BaseElement(hg.F(render))

        return hg.BaseElement(layout.ObjectFieldValue(self.column, rowvariable))

    class Meta:
        verbose_name = _("Column")
        verbose_name_plural = _("Columns")
        order_with_respect_to = "report"



from .export import *  # noqa
from .links import *  # noqa
from .model_helpers import *  # noqa
from .urls import *  # noqa


def get_all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in get_all_subclasses(c)]
    )



# Generated by Django 4.0.5 on 2022-08-19 06:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('triggers', '0003_auto_20220519_1018'),
    ]

    operations = [
        migrations.AddField(
            model_name='datachangetrigger',
            name='field',
            field=models.CharField(blank=True, help_text='Only trigger when a certain field has changed', max_length=255, verbose_name='Field'),
        ),
    ]



# Generated by Django 4.0.5 on 2022-07-24 12:58

from django.db import migrations, models

import basxbread.contrib.publicurls.models


class Migration(migrations.Migration):

    dependencies = [
        ('publicurls', '0002_alter_publicurl_options_publicurl_thankyou_text'),
    ]

    operations = [
        migrations.AlterField(
            model_name='publicurl',
            name='url',
            field=models.CharField(help_text='Internal URL', max_length=2048, validators=[basxbread.contrib.publicurls.models.validate_url], verbose_name='URL'),
        ),
    ]



from basxbread import menu
from basxbread.utils import Link, default_model_paths, reverse_model
from basxbread.views import AddView, BrowseView, EditView

from .models import Term, Vocabulary
from .views import TermsBrowseView, VocabularyBrowseView

urlpatterns = [
    *default_model_paths(
        Vocabulary,
        addview=AddView._with(fields=["name", "slug"]),
        editview=EditView._with(fields=["name"]),
        browseview=VocabularyBrowseView,
    ),
    *default_model_paths(
        Term,
        addview=AddView._with(fields=["term", "vocabulary", "slug"]),
        editview=EditView._with(fields=["term"]),
        browseview=TermsBrowseView._with(
            rowclickaction=BrowseView.gen_rowclickaction("edit", return_to_current=True)
        ),
    ),
]


menu.registeritem(
    menu.Item(
        Link(
            reverse_model(Vocabulary, "browse"),
            Vocabulary._meta.verbose_name_plural,
        ),
        menu.settingsgroup,
    )
)



import ast

import astor
import black
from django import forms
from django.core.exceptions import ValidationError
from django.db.models import Field
from django.utils.translation import gettext_lazy as _

from basxbread.utils import get_all_subclasses


def validate_choices(value):
    val = ast.literal_eval(value)
    if not isinstance(val, (tuple, list)):
        raise ValidationError(
            _("%(value)s is not a list or tuple"), params={"value": value}
        )
    for i in val:
        if len(i) != 2:
            raise ValidationError(
                _("%(value)s does not only contain 2-tuples"), params={"value": value}
            )


def valid_ordering(value):
    val = ast.literal_eval(value)
    for i in val:
        if not isinstance(i, str):
            raise ValidationError(
                _("%(value)s does not only contain list of fields"),
                params={"value": value},
            )
        i = i[1:] if i.startswith("-") else i
        if not i.isidentifier():
            raise ValidationError(
                _("%(value)s does not only contain list of fields"),
                params={"value": value},
            )


def validate_identifier(value):
    if not value.isidentifier():
        raise ValidationError(
            _("%(value)s is not a valid Python identifier"), params={"value": value}
        )


def validate_literal(value):
    try:
        ast.literal_eval(value)
    except Exception:
        raise ValidationError(
            _("%(value)s is not a valid Python literal"), params={"value": value}
        )


class ModelForm(forms.Form):
    name = forms.CharField(
        label=_("Model name"),
        help_text=_(
            "Python identifier for the model (no spaces and special characters). Only used internally."
        ),
        required=True,
        max_length=255,
        strip=True,
        validators=[validate_identifier],
    )
    verbose_name = forms.CharField(
        label=_("Verbose name"),
        help_text=_("Name that will be used as label for the model."),
        required=True,
        max_length=255,
        strip=True,
    )
    verbose_name_plural = forms.CharField(
        label=_("Verbose name plural"),
        help_text=_("Name that will be used as label for the model in plural form."),
        required=True,
        max_length=255,
        strip=True,
    )
    ordering = forms.CharField(
        label=_("Ordering"),
        help_text=_(
            "List of space separated fields that determine the default ordering for the objects of this model type"
        ),
        max_length=1024,
        strip=True,
        validators=[validate_literal, valid_ordering],
    )


def fieldtypes():
    return {
        f"{fieldtype.__module__}.{fieldtype.__qualname__}": fieldtype
        for fieldtype in get_all_subclasses(Field)
    }


def typechoices():
    return tuple(fieldtypes().items())


def typecoerce(value):
    return fieldtypes().get(value, None)


class FieldForm(forms.Form):
    name = forms.CharField(
        label=_("Field name"),
        help_text=_(
            "Python identifier for the field (no spaces and special characters). Only used internally."
        ),
        required=True,
        max_length=255,
        strip=True,
        validators=[validate_identifier],
    )
    type = forms.TypedChoiceField(
        choices=typechoices, coerce=typecoerce, empty_value=None
    )
    verbose_name = forms.CharField(
        label=_("Verbose name"),
        help_text=_("Name that will be used as label for the field."),
        required=False,
        max_length=255,
        strip=True,
    )
    null = forms.BooleanField(
        label=_("Null"),
        help_text=_("Allo this field to be null (not same as empty string)"),
        initial=False,
        required=False,
    )
    blank = forms.BooleanField(
        label=_("Blank"),
        help_text=_("Do not require this field to be filled out on forms"),
        initial=False,
        required=False,
    )
    editable = forms.BooleanField(
        label=_("Editable"),
        help_text=_("Allow this field to be edited via forms"),
        initial=True,
        required=False,
    )
    primary_key = forms.BooleanField(
        label=_("Primary key"),
        help_text=_("Make this field the primary key"),
        initial=False,
        required=False,
    )
    unique = forms.BooleanField(
        label=_("Unique"),
        help_text=_("Ensure this field is unique over all instances of the model"),
        initial=False,
        required=False,
    )
    help_text = forms.CharField(
        label=_("Help text"),
        help_text=_("Additional text to be displayed with the field"),
        max_length=1024,
        required=False,
    )
    default = forms.CharField(
        label=_("Default"),
        help_text=_("Default value for this field, must be a Python literal"),
        required=False,
        max_length=1024,
        validators=[validate_literal],
    )
    choices = forms.CharField(
        label=_("Choices"),
        help_text=_(
            "Iterable of 2-tuples with (value, label) literals, generates a select input"
        ),
        required=False,
        max_length=1024,
        validators=[validate_literal, validate_choices],
    )

    # Special, depending on field:
    # max_length: int (BinaryField, CharField, EmailField, FileField)
    # auto_now: bool (DateField, DateTimeField)
    # auto_now_add: bool (DateField, DateTimeField)
    # max_digits: int (DecimalField)
    # decimal_places: int (DecimalField)
    # upload_to: path (FileField)


def serialize(ast_module):
    return black.format_file_contents(
        astor.to_source(ast_module), fast=True, mode=black.FileMode()
    )


def field_ast2formdata(astfieldnode):
    pass


def is_variable_assignment(statement, name):
    return (
        isinstance(statement, ast.Assign)
        and len(statement.targets) == 1
        and statement.targets[0].id == name
    )


def is_single_param_call(statement):
    return (
        isinstance(statement, ast.Call)
        and len(statement.args) == 1
        and isinstance(statement.args[0], ast.Constant)
    )


def model_ast2formdata(astclassnode):
    modelname = astclassnode.name.lower()
    data = {"name": astclassnode.name}
    for statement in astclassnode.body:
        if isinstance(statement, ast.ClassDef) and statement.name == "Meta":
            for substatement in statement.body:
                if is_variable_assignment(substatement, "verbose_name"):
                    if is_single_param_call(substatement.value):
                        data["verbose_name"] = substatement.value.args[0].value
                    else:
                        data["verbose_name"] = substatement.value.value
                elif is_variable_assignment(substatement, "verbose_name_plural"):
                    if is_single_param_call(substatement.value):
                        data["verbose_name_plural"] = substatement.value.args[0].value
                    else:
                        data["verbose_name_plural"] = substatement.value.value
                elif is_variable_assignment(substatement, "ordering"):
                    data["ordering"] = astor.to_source(substatement.value).strip()

    return modelname, data


def parse(modelfilecontent):
    ret = {}
    module = ast.parse(modelfilecontent)
    for statement in module.body:
        if isinstance(statement, ast.ClassDef):
            modelname, data = model_ast2formdata(statement)
            ret[modelname] = data
    return ret



# Generated by Django 4.0.5 on 2022-08-15 09:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('document_templates', '0004_documenttemplatevariable_template'),
    ]

    operations = [
        migrations.AddField(
            model_name='documenttemplate',
            name='filename_template',
            field=models.TextField(blank=True, verbose_name='Filename template'),
        ),
    ]



from django.db import models
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _


class Vocabulary(models.Model):
    name = models.CharField(_("Name"), max_length=255, unique=True)
    slug = models.SlugField(
        _("Slug"),
        unique=True,
        help_text=_("slug is human-readable, to make referencing easier"),
    )

    def __str__(self):
        return self.name

    class Meta:
        ordering = ["name"]
        verbose_name = _("Vocabulary")
        verbose_name_plural = _("Vocabularies")


class Term(models.Model):
    vocabulary = models.ForeignKey(Vocabulary, null=False, on_delete=models.CASCADE)
    vocabulary.verbose_name = _("Vocabulary")
    term = models.CharField(_("Term"), max_length=255)
    slug = models.CharField(_("Slug"), max_length=255, unique=True, blank=True)

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = self.vocabulary.slug + "__" + slugify(self.term)
        super().save(*args, **kwargs)

    def __str__(self):
        return self.term

    class Meta:
        verbose_name = _("Term")
        verbose_name_plural = _("Terms")
        order_with_respect_to = "vocabulary"
